H_PARAMS = [
    "b1",
    "b2",
    "batch_size",
    "channel_exp_base",
    "down_stride",
    "dilation_factor",
    "fc_repr_dim",
    "fc_task_dim",
    "fc_do",
    "first_kernel_expansion",
    "first_stride_expansion",
    "kernel_width",
    "l1",
    "lr",
    "split_mlp_num_splits",
    "mg_num_experts",
    "mixing_alpha",
    "lr_schedule",
    "model_type",
    "na_augment_perc",
    "na_augment_prob",
    "optimizer",
    "rb_do",
    "layers",
    "sa",
    "warmup_steps",
    "wd",
]
